//============================================================================
// Name        : genvisitor.cpp
// Author      : G Lipari
// Version     :
// Copyright   : OOSD
// Description : Hello World in C++, Ansi-style
//============================================================================

#include <iostream>
#include <loki/Visitor.h>

using namespace std;
using namespace Loki;

class Node : public BaseVisitable<> {
protected:
	string type;
public:
	Node (const string &n) : type(n) {}
	virtual ~Node() {}

	string getType() const { return type; }
	virtual bool lvalue() const = 0;
	virtual bool isLeaf() { return true; }

	LOKI_DEFINE_VISITABLE()
};

class CompositeNode : public Node {
public:
	CompositeNode(const string &n) : Node(n) {}
	bool isLeaf() { return false; }

	virtual Node * getLeft() = 0;
	virtual Node * getRight() = 0;

	LOKI_DEFINE_VISITABLE()
};

class VariableRefNode : public Node {
	string name;
public:
	VariableRefNode(const string &n) : Node("var"), name(n) {}
	bool lvalue() const { return true; }
	string getName() { return name; }

	LOKI_DEFINE_VISITABLE()
};

class NumberNode : public Node {
	int number;
public:
	NumberNode(int i) : Node("number"), number(i) {}
	bool lvalue() const { return false; }
	int getValue() { return number; }

	LOKI_DEFINE_VISITABLE()
};

class BinaryOperatorNode : public CompositeNode {
	Node *left;
	Node *right;
	string op;
public:
	BinaryOperatorNode(const string &opname, const string &n) :
		CompositeNode(opname), left(0), right(0), op(n) {}

	void setLeft(Node *v) { left = v; }
	void setRight(Node *v) { right = v; }
	bool lvalue() const { return false; }
	Node * getLeft() { return left; }
	Node * getRight() { return right; }

	LOKI_DEFINE_VISITABLE()
};

class AssignmentNode : public BinaryOperatorNode {
public:
	AssignmentNode() : BinaryOperatorNode("assignment", "=") {}

	LOKI_DEFINE_VISITABLE()
};

class SumNode : public BinaryOperatorNode {
public:
	SumNode() : BinaryOperatorNode("sum", "+") {}

	LOKI_DEFINE_VISITABLE()
};

/* --------------------------------------- */

class GenerateCode : public BaseVisitor,
	public Visitor<LOKI_TYPELIST_4(VariableRefNode, NumberNode, AssignmentNode, SumNode)>
{
public:
	void Visit(VariableRefNode &n) {
		cout << n.getName();
	}
	void Visit(NumberNode &n) {
		cout << n.getValue();
	}
	void Visit(AssignmentNode &n) {
		cout << " = ";
	}
	void Visit(SumNode &n) {
		cout << " + ";
	}
};

void in_order_visit(Node *n, BaseVisitor *v)
{
	if (n->isLeaf()) n->Accept(*v);
	else {
		CompositeNode *cn = dynamic_cast<CompositeNode *>(n);
		if (cn == 0) throw "Error, should be a composite";
		if (cn->getLeft()) in_order_visit(cn->getLeft(), v);
		cn->Accept(*v);
		if (cn->getRight()) in_order_visit(cn->getRight(), v);
	}
}

void gen_code(Node *n)
{
	GenerateCode visitor;
	in_order_visit(n, &visitor);
}

int main() {
	AssignmentNode an;
	VariableRefNode vr("x");
	NumberNode n1(10);
	NumberNode n2(20);
	SumNode sn;

	sn.setLeft(&n1);
	sn.setRight(&n2);

	an.setRight(&sn);
	an.setLeft(&vr);

	//type_check(&an);
	gen_code(&an);

	return 1;
}
